/*
 * Copyright (c) 2010-2011 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.dua.aggrlzzsdtv.AggrLzzsDTV
 *
 * de.inovat.dua.aggrlzzsdtv.AggrLzzsDTV is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.dua.aggrlzzsdtv.AggrLzzsDTV is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.dua.aggrlzzsdtv.AggrLzzsDTV; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.inovat.dua.aggrlzzsdtv;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

import java.text.SimpleDateFormat;
import java.util.*;

import static java.util.Calendar.HOUR_OF_DAY;
import static java.util.GregorianCalendar.MILLISECOND;
import static java.util.GregorianCalendar.MINUTE;
import static java.util.GregorianCalendar.SECOND;

//~ KLASSEN ===================================================================

/**
 * Klasse, die die Aggregation der Langzeitzählstellen (LZZS) durchführt.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class AggrLzzsDTV implements StandardApplication, ClientReceiverInterface, ClientSenderInterface {

    /** WarteZeit in ms, bis die Auftragsliste erneut geprüft wird, nachdem keinerlei Aufträge mehr vorlagen * */
    private static final long WARTEZEIT_AUF_NAECHSTEN_AUFTRAG = 60 * 1000;  // 1 Minute

    /** DebugLogger fuer Debug-Ausgaben. */
    private static Debug debug;

    //~ FELDER ================================================================

    private long _bisZeitPunkt;
    private long _bisZeitPunktAufrufParameter;

    /** Datenverteilerverbindung */
    private ClientDavInterface _dav;
    private boolean            _einmaligeNachBerechnung;
    private boolean            _einmaligeNachBerechnungAufrufParameter;
    private String             _grenzenGeschwindigKeitsKlassenBPAAufrufParameter;
    private String             _grenzenGeschwindigKeitsKlassenLVOAufrufParameter;
    private String             _grenzenGeschwindigKeitsKlassenSGVAufrufParameter;
    private int                _intervallDauerInStunden;
    private String             _intervallDauerInStundenAufrufParameter;
    private boolean            _isAggregationAusErfasstenDaten;
    private boolean            _isAggregationAusErfasstenDatenAufrufParameter;
    private boolean            _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall;
    private boolean            _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervallAufrufParameter;

    /** Die Konfiguration des DaV. */
    private DataModel            _konfiguration;
    private String               _listePidsAufrufParameter;
    private int                  _minimaleWarteZeitInStunden;
    private String               _minimaleWarteZeitInStundenAufrufParameter;
    private int                  _nachberechnungInTagen;
    private String               _nachberechnungInTagenAufrufParameter;
    private String               _pidVerwaltungsObjektAggrLzzsAufrufParameter;
    private SystemObject         _verwaltungsObjektAggrLzzs;
    private long                 _vonZeitPunkt;
    private long                 _vonZeitPunktAufrufParameter;
    private List<String>         _listePidsLzzs               = new ArrayList<String>();
    private List<SystemObject>   _listeLzzsSystemObjekte      = new ArrayList<SystemObject>();
    private List<String>         _listePidsLzzsNichtVorhanden = new ArrayList<String>();
    private Map<String, VewLZZS> _mapVewLZZS                  = new HashMap<String, VewLZZS>();

    //~ METHODEN ==============================================================

    /**
     * Main-Methode der SWE Aggregation LZZS DTV.
     *
     * @param args Aufrufargumente der Applikation.
     */
    public static void main(String[] args) {
        AggrLzzsDTV aggrLzzsDTV = new AggrLzzsDTV();

        try {
            StandardApplicationRunner.run(aggrLzzsDTV, args);

            // Debug erst hier initialisieren, da erst ab hier verfügbar !!
            debug = Debug.getLogger();
            aggrLzzsDTV.auswertungAufrufParameterPidVerwaltungsObjekt();
            aggrLzzsDTV.auswertungAufrufParameterListePids();
            aggrLzzsDTV.auswertungAufrufParameterVonBisIntervallNachberechnung();
            aggrLzzsDTV.auswertungAufrufParameterGrenzenXXXParameter();
            aggrLzzsDTV.initialisierung();
            debug.config(aggrLzzsDTV.ausgabeKonfigurationsZusammenfassung());
        }
        catch (RuntimeException e) {
            e.printStackTrace();
            System.exit(-1);
        }

        aggrLzzsDTV.run();
    }

    /**
     * Gibt Informationen zur Konfiguration aus. Ist nur zu Anlaysezwecken bzw. zur zusammenfassenden Ausgabe der
     * Konfigurationsphase dieser Applikatione interessant.
     *
     * @return Informationen zur Konfiguration (Aufbau der Ausgabe kann sich ändern).
     */
    private String ausgabeKonfigurationsZusammenfassung() {
        StringBuffer sb = new StringBuffer();

        // Listen der SystemObjekte sortieren...
        Collections.sort(_listeLzzsSystemObjekte, new Comparator<SystemObject>() {
            public int compare(SystemObject o1, SystemObject o2) {
                return o1.getPid().compareTo(o2.getPid());
            }
        });
        Collections.sort(_listePidsLzzsNichtVorhanden);
        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("START: Konfigurationszusammenfassung für AggrLzzsDTV"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -pidVerwaltungsObjektAggrLzzsDtv=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _pidVerwaltungsObjektAggrLzzsAufrufParameter));

        if (_verwaltungsObjektAggrLzzs != null) {
            sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _verwaltungsObjektAggrLzzs.getPid()));
        } else {
            sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _verwaltungsObjektAggrLzzs));
            sb.append(String.format("Verwaltungsobjekt wird nicht weiter verwendet.%nExterne Steuerung zur Laufzeit ist damit nicht mehr aktiviert.%n"));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Im Aufrufparameter angegebene PIDs, zu denen keine SystemObjekte existieren"));
        sb.append(String.format("%n====================================================================================%n"));

        int i = 0;

        for (String pid : _listePidsLzzsNichtVorhanden) {
            sb.append(String.format("%04d: %s%n", ++i, pid));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Liste der PIDs der verwendeten Systemobjekte"));
        sb.append(String.format("%n====================================================================================%n"));

        int j = 0;

        for (SystemObject so : _listeLzzsSystemObjekte) {
            sb.append(String.format("%04d: %s%n", ++j, so.getPidOrNameOrId()));
        }

        SimpleDateFormat datumsFormatierer = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -intervallDauerInStunden=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _intervallDauerInStundenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _intervallDauerInStunden));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -nachberechnungInTagen=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _nachberechnungInTagenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _nachberechnungInTagen));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -vonZeitPunkt=..."));
        sb.append(String.format("%n====================================================================================%n"));

        Date datum = new Date(_vonZeitPunktAufrufParameter);

        datum.toString();
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktAufrufParameter))));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunkt))));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -bisZeitPunkt=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", datumsFormatierer.format(new Date(_bisZeitPunktAufrufParameter))));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", datumsFormatierer.format(new Date(_bisZeitPunkt))));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -einmaligeNachBerechnung=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _einmaligeNachBerechnungAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _einmaligeNachBerechnung));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -aggregationAusErfasstenDaten=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _isAggregationAusErfasstenDatenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _isAggregationAusErfasstenDaten));

        if (!_isAggregationAusErfasstenDaten) {
            sb.append(String.format("%nAggregation erfolgt auf Basis der reimportierten BASt-Daten.%nALLE AGGREGATIONSAUFTRÄGE MÜSSEN MANUELL AUSGEFÜHRT WERDEN.%n"));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -qWerteAusReimportAtgVerkehrsDatenLangzeitIntervall=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervallAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall));

        if (!_isAggregationAusErfasstenDaten && _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall) {
            sb.append(String.format("%nVor der Aggregation werden die 8+1-q-Werte aus aus den reimportierten BASt V2004-Werten (atg.verkehrsDatenLangZeitIntervall, asp.messWertErsetzung) übernommen.%n"));
        }

        if (_isAggregationAusErfasstenDaten && _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall) {
            sb.append(String.format("%nParameter wird nicht berücksichtigt, da -aggregationAusErfasstenDaten=ja.%n"));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -minimaleWarteZeitInStunden=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _minimaleWarteZeitInStundenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _minimaleWarteZeitInStunden));

        Parameter p = Parameter.getInstanz();

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -grenzenGeschwindigKeitsKlassenLVO=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _grenzenGeschwindigKeitsKlassenLVOAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", p.ausgabe(p.getGrenzenLVO())));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -grenzenGeschwindigKeitsKlassenSGV=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _grenzenGeschwindigKeitsKlassenSGVAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", p.ausgabe(p.getGrenzenSGV())));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -grenzenGeschwindigKeitsKlassenBPA=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _grenzenGeschwindigKeitsKlassenBPAAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", p.ausgabe(p.getGrenzenBPA())));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Auftragsliste initial..."));
        sb.append(String.format("%n====================================================================================%n"));

        int k = 0;

        for (Auftrag auftrag : AuftragsListe.getInstanz().holeKopieAuftragsListe()) {
            sb.append(String.format("%04d: %s%n", ++k, auftrag.toString()));
        }

        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("ENDE: Konfigurationszusammenfassung für AggrLzzsDTV"));
        sb.append(String.format("%n####################################################################################%n"));

        return sb.toString();
    }

    /** Werte die Aufrufparameter final aus und setzt die davon abhängigen Parameter. */
    private void auswertungAufrufParameterGrenzenXXXParameter() {
        Parameter p = Parameter.getInstanz();

        p.initialisiereGrenzenLVO(_grenzenGeschwindigKeitsKlassenLVOAufrufParameter);
        p.initialisiereGrenzenSGV(_grenzenGeschwindigKeitsKlassenSGVAufrufParameter);
        p.initialisiereGrenzenBPA(_grenzenGeschwindigKeitsKlassenBPAAufrufParameter);
    }

    /**
     * Wertet die Aufrufparameter
     * <p/>
     * -listePids
     * <p/>
     * wie folgt aus:
     * <p/>
     * Es werden alle Systemobjekte vom Typ <tt>typ.langZeitZählstelle</tt> ermittelt und dabei diejenigen für die weiteren
     * Berechnungen übernommen, deren Pid in der Liste der Pids aufgeführt ist (bzw. alle, falls Aufruf mit
     * <tt>-listePids=alle</tt> erfolgt). Wird kein passendes Objekt gefunden, wird das Programm mit einer Fehlermeldung
     * abgebrochen.
     */
    private void auswertungAufrufParameterListePids() {
        try {

            // Pids der zu berücksichtigenden Objekte ermitteln
            _listePidsLzzs = Arrays.asList(_listePidsAufrufParameter.split("(:|,|;)"));

            // Alle möglichen SystemObjekte ermittlen...
            List<SystemObject> listeAlleLzzsSystemObjekte = _konfiguration.getType("typ.langZeitZählstelle").getElements();

            // ... und jetzt Liste mit zu verwendenden aufbauen (falls nicht "alle" verwendet werden sollen) ...
            boolean isAlle = ((_listePidsLzzs.size() == 1) && _listePidsLzzs.get(0).equals("alle"));

            for (SystemObject so : listeAlleLzzsSystemObjekte) {
                if (isAlle || _listePidsLzzs.contains(so.getPid())) {
                    _listeLzzsSystemObjekte.add(so);
                }
            }

            // .. und für Konfigurationszusammenfassung Liste der nicht verwendeten PID's aus dem Aufrufparameter ermitteln
            List<String> listePidsVerwendet = new ArrayList<String>();

            for (SystemObject so : _listeLzzsSystemObjekte) {
                listePidsVerwendet.add(so.getPid());
            }

            for (String pid : _listePidsLzzs) {
                if (!listePidsVerwendet.contains(pid)) {
                    _listePidsLzzsNichtVorhanden.add(pid);
                }
            }
        }
        catch (Exception e) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZählstelle ermittelt werden!. Programm wird beendet.", e);

            throw new RuntimeException();
        }

        if (_listeLzzsSystemObjekte.size() == 0) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZählstelle ermittelt werden!. Programm wird beendet.");

            throw new RuntimeException();
        }
    }

    /**
     * Wertet den Aufrufparameter
     * <p/>
     * -pidVerwaltungsObjektAggrLzzsDtv
     * <p/>
     * wie folgt aus:
     * <p/>
     * Ist der Wert <code>automatisch</code>, so wird das Objekt mit dem Typ <code>typ.aggrLzzsDtv</code> und der Pid
     * <code>aggrLzzsDtv.vewStandard1</code> verwendet, sonst das Objekt mit der im Aufrufparameter übergebenen Pid..
     */
    private void auswertungAufrufParameterPidVerwaltungsObjekt() {
        try {
            if (_pidVerwaltungsObjektAggrLzzsAufrufParameter.equals("automatisch")) {
                _verwaltungsObjektAggrLzzs = _konfiguration.getObject("aggrLzzsDtv.vewStandard1");
            } else {
                _verwaltungsObjektAggrLzzs = _konfiguration.getObject(_pidVerwaltungsObjektAggrLzzsAufrufParameter);
            }

            if (_verwaltungsObjektAggrLzzs != null) {
                if (!_verwaltungsObjektAggrLzzs.isOfType("typ.aggrLzzsDtv")) {
                    debug.warning(String.format("VerwaltungsObjekt für Aggregation Langzeitzählstellen [-pidVerwaltungsObjektAggrLzzsDtv=%s] ist nicht vom Typ [typ.aggrLzzsDtv] sondern vom Typ [%s].%nVerwaltungsobjekt wird nicht weiter verwendet. Externe Steuerung zur Laufzeit ist damit nicht mehr aktiviert.", _pidVerwaltungsObjektAggrLzzsAufrufParameter, _verwaltungsObjektAggrLzzs.getType().getPid()));
                }
            } else {
                debug.warning(String.format("VerwaltungsObjekt für Aggregation Langzeitzählstellen [-pidVerwaltungsObjektAggrLzzsDtv=%s] konnte nicht ermittelt werden.%nVerwaltungsobjekt wird nicht weiter verwendet. Externe Steuerung zur Laufzeit ist damit nicht mehr aktiviert.", _pidVerwaltungsObjektAggrLzzsAufrufParameter));
            }
        }
        catch (Exception e) {
            debug.warning(String.format("VerwaltungsObjekt für Aggregation Achslast [-pidVerwaltungsObjektAggrLzzsDtv=%s] konnte nicht ermittelt werden.%nVerwaltungsobjekt wird nicht weiter verwendet. Externe Steuerung zur Laufzeit ist damit nicht mehr aktiviert.", _pidVerwaltungsObjektAggrLzzsAufrufParameter), e);
        }
    }

    /**
     * Wertet die Aufrufparameter
     * <p/>
     * -vonZeitPunkt
     * <p/>
     * -bisZeitPunkt
     * <p/>
     * -intervallDauerInStunden
     * <p/>
     * -nachberechnungInTagen
     * <p/>
     * -einmaligeNachBerechnung
     * <p/>
     * wie folgt aus:
     * <p/>
     * <tt>-intervallDauerInStunden</tt>: Wird auf gültige Werte (1,2,3,4,6,8,12,24) geprüft und bei ungültigem Wert auf 1
     * Stunde gesetzt.
     * <p/>
     * <tt>-nachberechnungInTagen</tt>: Wird auf Gültigkeit getestet (ganze Zahl) und bei ungültigem Wert auf 0 Tage (keine
     * Nachberechnung) gesetzt.
     * <p/>
     * <tt>-vonZeitPunkt</tt>: Wird auf gültigen Intervallanfangszeitpunkt abhängig vom Intervalltyp und
     * <tt>-nachberechnungInTagen</tt> auf denn nächsten, davorliegenden Intervallbeginn gesetzt.
     * <p/>
     * Beispiel: Ist <tt>-vonZeitPunkt=04.11.2010 11:51:01,956</tt>, betrachtets Intervall DTV_TAG und
     * <tt>-nachberechnungInTagen=7</tt>, so wird der Zeitpunkt auf <tt>28.10.2010 00:00:00,000</tt> gesetzt.
     * <p/>
     * <tt>-bisZeitPunkt</tt>: Wird auf gültigen Intervallanfangszeitpunkt abhängig vom Intervalltyp auf denn nächsten,
     * dahinterliegenden Intervallbeginn gesetzt. Beispiel: Ist Intervalltyp DTV_TAG und <tt>-bisZeitPunkt=04.11.2010
     * 11:51:01,956</tt>, so wird der Zeitpunkt auf <tt>05.11.2010 00:00:00,000</tt> gesetzt. Liegt der so ermittelte
     * bisZeitPunkt nicht mindestens ein Intervall hinter dem vonZeitPunkt, wird er automatisch dahingehend korrigiert. Der
     * bisZeitPunkt wird nur berücksichtigt, wenn der Parameter <tt>-einmaligeNachBerechnung=ja</tt> ist. Ist der Parameter
     * <tt>-einmaligeNachBerechnung=nein</tt>, so wird ab dem Startzeitpunkt bis zum Beenden des Prozesses die Aggregation
     * durchgeführt.
     */
    private void auswertungAufrufParameterVonBisIntervallNachberechnung() {

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -nachberechnungInTagen
        // ------------------------------------------------------------------------------------------------------
        try {
            _nachberechnungInTagen = new Integer(_nachberechnungInTagenAufrufParameter);

            if (_nachberechnungInTagen < 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _nachberechnungInTagen = 0;
            debug.warning(String.format("Aufrufparameter [-nachberechnungInTagen=%s] ist ungültig. Wert wurde für weiter Berechnung auf [-nachberechnungInTagen=%d] gesetzt.", _nachberechnungInTagenAufrufParameter, _nachberechnungInTagen));
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -intervallDauerInStunden
        // ------------------------------------------------------------------------------------------------------
        try {
            Set<Integer> gueltigeWerte = new HashSet<Integer>();

            gueltigeWerte.add(1);
            gueltigeWerte.add(2);
            gueltigeWerte.add(3);
            gueltigeWerte.add(4);
            gueltigeWerte.add(6);
            gueltigeWerte.add(8);
            gueltigeWerte.add(12);
            gueltigeWerte.add(24);
            _intervallDauerInStunden = new Integer(_intervallDauerInStundenAufrufParameter);

            if (!gueltigeWerte.contains(_intervallDauerInStunden)) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _intervallDauerInStunden = 1;
            debug.warning(String.format("Aufrufparameter [-intervallDauerInStunden=%s] ist ungültig (nur 1,2,3,4,6,8,12,24 zulässig). Wert wurde für weiter Berechnung auf [-intervallDauerInStunden=%d] gesetzt.", _intervallDauerInStundenAufrufParameter, _intervallDauerInStunden));
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        _vonZeitPunkt = ermittleIntervallBeginn(_vonZeitPunktAufrufParameter, IntervallDauer.DTV_TAG, _nachberechnungInTagen);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bisZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        _bisZeitPunkt = ermittleIntervallEnde(_bisZeitPunktAufrufParameter, IntervallDauer.DTV_TAG);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -einmaligeNachBerechnung
        // ------------------------------------------------------------------------------------------------------
        _einmaligeNachBerechnung = _einmaligeNachBerechnungAufrufParameter;

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -aggregationAusErfasstenDaten
        // ------------------------------------------------------------------------------------------------------
        _isAggregationAusErfasstenDaten = _isAggregationAusErfasstenDatenAufrufParameter;

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -qWerteAusReimportAtgVerkehrsDatenLangzeitIntervall
        // ------------------------------------------------------------------------------------------------------
        _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall = _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervallAufrufParameter;

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStunden
        // ------------------------------------------------------------------------------------------------------
        try {
            _minimaleWarteZeitInStunden = new Integer(_minimaleWarteZeitInStundenAufrufParameter);

            if (_minimaleWarteZeitInStunden <= 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _minimaleWarteZeitInStunden = 1;
            debug.warning(String.format("Aufrufparameter [-minimaleWarteZeitInStunden=%s] ist ungültig. Wert wurde für weiter Berechnung auf [-minimaleWarteZeitInStunden=%d] gesetzt.", _minimaleWarteZeitInStundenAufrufParameter, _minimaleWarteZeitInStunden));
        }
    }

    /**
     * Sendesteuerung des Datenverteilers an die Applikation. Diese Methode muss von der Applikation implementiert werden,
     * um den Versand von Daten zu starten bzw. anzuhalten. Der Datenverteiler signalisiert damit einer Quelle oder einem
     * Sender dass mindestens ein Abnehmer bzw. kein Abnehmer mehr für die zuvor angemeldeten Daten vorhanden ist. Die
     * Quelle wird damit aufgefordert den Versand von Daten zu starten bzw. zu stoppen.
     *
     * @param object          Das in der zugehörigen Sendeanmeldung angegebene Objekt, auf das sich die Sendesteuerung
     *                        bezieht.
     * @param dataDescription Beschreibende Informationen zu den angemeldeten Daten auf die sich die Sendesteuerung
     *                        bezieht.
     * @param state           Status der Sendesteuerung. Kann einen der Werte <code>START_SENDING</code>,
     *                        <code>STOP_SENDING</code>, <code>STOP_SENDING_NO_RIGHTS</code>,
     *                        <code>STOP_SENDING_NOT_A_VALID_SUBSCRIPTION</code> enthalten.
     *
     * @see #START_SENDING
     * @see #STOP_SENDING
     * @see #STOP_SENDING_NO_RIGHTS
     * @see #STOP_SENDING_NOT_A_VALID_SUBSCRIPTION
     */
    @Override
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {

        // tut nicht, da Sendesteuerung nicht unterstützt wird
    }

    private long ermittleAnzMonate(Calendar intervallBeginnMonat, Calendar intervallEndeMonat) {

        // Fall A: Anfang und Ende liegen im selben Jahr
        if (intervallEndeMonat.get(Calendar.YEAR) == intervallBeginnMonat.get(Calendar.YEAR)) {
            return intervallEndeMonat.get(Calendar.MONTH) - intervallBeginnMonat.get(Calendar.MONTH);
        }

        // Fall B: Anfang und Ende liegen in verschiedenen Jahren
        // Anzahl der Monate bis Jahresende im ersten Jahr
        // plus Anzahl der Monate im letzten Jahr
        // plus 12 * Anzahl der vollen Jahre
        long anzGanzeJahre       = intervallEndeMonat.get(Calendar.YEAR) - intervallBeginnMonat.get(Calendar.YEAR) - 1;
        long anzMonatErstesJahr  = 11 - intervallBeginnMonat.get(Calendar.MONTH);
        long anzMonatLetztesJahr = intervallEndeMonat.get(Calendar.MONTH) + 1;

        return anzGanzeJahre * 12 + anzMonatErstesJahr + anzMonatLetztesJahr;
    }

    /**
     * Ermittelt den IntervallBeginn ausgehend von einem Zeitpunk, der Intervalldauer und der Nachberechnungszeit
     *
     * @param spaetesterIntervallBeginn Anfangszeitpunkt in ms seit 1.1.1970, ab dem spätestens die Daten berechnet werden
     *                                  sollen. Dieser Wert wird dann noch unter Berücksichtigung des Parameters
     *                                  <tt>nachberechnungInTagen</tt> soweit in die Vergangenheit verschoben, dass genau
     *                                  ein gültiger Intervallbeginn (Tagesbeginn) ermittelt wird.
     * @param intervallDauer            Intervalldauer, die gerechnet werden soll.
     * @param nachberechnungInTagen     Ggf. Zeit die vor dem <tt>vonZeitPunkt</tt> zusätzlich noch berücksichtigt werden
     *                                  soll.
     *
     * @return Intervallbeginn (GENAU!!), der vor dem Zeitpunkt <tt>vonZeitPunkt</tt> liegt
     */
    private long ermittleIntervallBeginn(long spaetesterIntervallBeginn, IntervallDauer intervallDauer, int nachberechnungInTagen) {
        Calendar vonZeitPunkt = new GregorianCalendar();

        // vonZeitPunkt darf spätestens sein...
        vonZeitPunkt.setTimeInMillis(spaetesterIntervallBeginn - (nachberechnungInTagen * 24 * 60 * 60 * 1000));

        // ... und muss jetzt auf gültigen Intervallzeitpunkt DAVOR abgebildet werden...
        vonZeitPunkt.set(MILLISECOND, 0);
        vonZeitPunkt.set(SECOND, 0);
        vonZeitPunkt.set(MINUTE, 0);
        vonZeitPunkt.set(HOUR_OF_DAY, 0);

        switch (intervallDauer) {
        case DTV_TAG :

            // Ist schon ok
            break;

        case DTV_MONAT :

            // Tag auf 1 setzen
            vonZeitPunkt.set(Calendar.DAY_OF_MONTH, 1);

            break;

        case DTV_QUARTAL :

            // Tag auf 1 setzen
            vonZeitPunkt.set(Calendar.DAY_OF_MONTH, 1);

            // Monat auf Quartalsanfang korrigieren
            int monat = vonZeitPunkt.get(Calendar.MONTH);

            switch (monat) {
            case 0 :
            case 1 :
            case 2 :
                vonZeitPunkt.set(Calendar.MONTH, 0);

                break;

            case 3 :
            case 4 :
            case 5 :
                vonZeitPunkt.set(Calendar.MONTH, 3);

                break;

            case 6 :
            case 7 :
            case 8 :
                vonZeitPunkt.set(Calendar.MONTH, 6);

                break;

            case 9 :
            case 10 :
            case 11 :
                vonZeitPunkt.set(Calendar.MONTH, 9);

                break;
            }

            break;

        case DTV_JAHR :
        case MSV_JAHR :

            // Tag auf 1 setzen
            vonZeitPunkt.set(Calendar.DAY_OF_MONTH, 1);

            // Monat auf 1 setzen
            vonZeitPunkt.set(Calendar.MONTH, 0);

            break;
        }

        return vonZeitPunkt.getTimeInMillis();
    }

    /**
     * Ermittelt das IntervallEnde ausgehend von einem Zeitpunkt und der Intervalldauer
     *
     * @param fruehestesIntervallEnde Endzeitpunkt, bis zu dem mindestens gerechnet werden muss.
     * @param intervallDauer          Intervalldauer, die gerechnet werden soll.
     *
     * @return Intervallende (AUSSCHLIESSLICH!!), also der Intervallbeginn des nächsten Intervalls
     */
    private long ermittleIntervallEnde(long fruehestesIntervallEnde, IntervallDauer intervallDauer) {
        Calendar bisZeitPunkt = new GregorianCalendar();

        bisZeitPunkt.setTimeInMillis(ermittleIntervallBeginn(fruehestesIntervallEnde, intervallDauer, 0));

        switch (intervallDauer) {
        case DTV_TAG :

            // Zum Intervallanfang 24 Stunden hinzufügen
            bisZeitPunkt.add(Calendar.HOUR_OF_DAY, 24);

            break;

        case DTV_MONAT :

            // Zum Intervallanfang 1 Monat hinzufügen
            bisZeitPunkt.add(Calendar.MONTH, 1);

            break;

        case DTV_QUARTAL :

            // Zum Intervallanfang 3 Monat hinzufügen
            bisZeitPunkt.add(Calendar.MONTH, 3);

            break;

        case DTV_JAHR :
        case MSV_JAHR :

            // Zum Intervallanfang 1 Jahr hinzufügen
            bisZeitPunkt.add(Calendar.YEAR, 1);

            break;
        }

        return bisZeitPunkt.getTimeInMillis();
    }

    /**
     * Folgeauftrag für den aktuell übergebenen Auftrag berechnen.
     * <p/>
     * Der Folgeauftrag liegt genau eine Intervalllänge später, sofern dieser Zeitpunkt nach innerhalb des insgesamt zu
     * berücksichtigenden Zeitbereichs liegt (Aufrufparameter <tt>-vonZeitPunkt, -bisZeitPunkt,
     * -einmaligeNachBerechnung</tt>.
     *
     * @param auftrag Der aktuelle Auftrag, für den ein Folgeauftrag berechnet werden soll.
     *
     * @return Folgeauftrag für den aktuell als Parameter übergebenen Auftrag oder <code>null</code>, wenn kein
     *         Folgeauftrag mehr existiert.
     */
    private Auftrag erstelleFolgeAuftrag(Auftrag auftrag) {

        // Prüfen, ob Folgeauftrag überhaupt noch berechnet werden soll, weil Ursprungsauftrag Einmalauftrag war:
        if (auftrag.isEinmalAuftrag()) {
            return null;
        }

        // Sonst Folgeauftrag zusammenstellen und prüfen, ob dieser überhaupt noch ausgeführt werden soll.
        Auftrag folgeAuftrag = new Auftrag(auftrag);

        // Prüfen, ob Folgeauftrag überhaupt noch berechnet werden soll, weil Gesamtzeitraum nur einmalig berechnet werden soll:
        if (_einmaligeNachBerechnung && (folgeAuftrag.getIntervallEnde() > _bisZeitPunkt)) {
            return null;
        }

        return folgeAuftrag;
    }

    /** Initialisierung von AggrLzzsDTV */
    private void initialisierung() {

        // VewLZZS-Objekte konstruieren
        for (SystemObject lzzs : _listeLzzsSystemObjekte) {
            VewLZZS vewLZZS = new VewLZZS(_dav, lzzs, _intervallDauerInStunden, _isAggregationAusErfasstenDaten, _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervall);

            _mapVewLZZS.put(lzzs.getPid(), vewLZZS);
        }

        // ---------------------------------------------------------------------------------
        // Initiales setzen der Auftragsliste auf Basis der VewLZZS und der Aurufparameter
        //
        // Es werden nur dann initial Daueraufträge eingetragen, wenn der Aufrufparameter
        // -aggregationAusErfasstenDaten = ja steht (d. h. die Aggregation erfolgt
        // auf Basis der erfassten Daten und NICHT auf Basis der Reimportierten BASt-Daten.
	    // Bei Reimportierten Daten erfolgt die Auftragserteilung ausschließlich über die
	    // Datenverteilerschnittstelle zur Auftragssteuerung, also i.d.R. manuell durch
	    // den Anwender nach Durchführung des Import der notwendigen BASt-Daten V2004 und V2007
        // ---------------------------------------------------------------------------------
        if (_isAggregationAusErfasstenDaten) {
            for (VewLZZS vewLZZS : _mapVewLZZS.values()) {

                // Zeitpunkte für den TAG bestimmen
                long intervallBeginnTag       = _vonZeitPunkt;
                long intervallEndeTag         = ermittleIntervallEnde(intervallBeginnTag, IntervallDauer.DTV_TAG);
                long ausfuehrungsZeitpunktTag = intervallEndeTag + (_minimaleWarteZeitInStunden * 60 * 60 * 1000);

                // Zeitpunkte für den MONAT bestimmen
                long intervallBeginnMonat = ermittleIntervallBeginn(intervallBeginnTag, IntervallDauer.DTV_MONAT, 0);
                long intervallEndeMonat   = ermittleIntervallEnde(intervallBeginnMonat, IntervallDauer.DTV_MONAT);

                // Verschiebung um einen Monat, beim ersten Start erst mit dem Folgemonat beginnen !
                // intervallBeginnMonat = intervallEndeMonat;
                // intervallEndeMonat = ermittleIntervallEnde(intervallBeginnMonat, IntervallDauer.MONAT);
                long ausfuehrungsZeitpunktMonat = intervallEndeMonat + ((_minimaleWarteZeitInStunden + 1) * 60 * 60 * 1000);

                // Zeitpunkte für das QUARTAL bestimmen
                long intervallBeginnQuartal = ermittleIntervallBeginn(intervallBeginnTag, IntervallDauer.DTV_QUARTAL, 0);
                long intervallEndeQuartal   = ermittleIntervallEnde(intervallBeginnQuartal, IntervallDauer.DTV_QUARTAL);

                // Verschiebung um ein Quartal, beim ersten Start erst mit dem Fogequartal beginnen !
                // intervallBeginnQuartal = intervallEndeQuartal;
                // intervallEndeQuartal = ermittleIntervallEnde(intervallBeginnQuartal, IntervallDauer.QUARTAL);
                long ausfuehrungsZeitpunktQuartal = intervallEndeQuartal + ((_minimaleWarteZeitInStunden + 2) * 60 * 60 * 1000);

                // Zeitpunkte für das JAHR bestimmen
                long intervallBeginnJahr = ermittleIntervallBeginn(intervallBeginnTag, IntervallDauer.DTV_JAHR, 0);
                long intervallEndeJahr   = ermittleIntervallEnde(intervallBeginnJahr, IntervallDauer.DTV_JAHR);

                // Verschiebung um ein Jahr, beim ersten Start erst mit dem Fogejahr beginnen !
                // intervallBeginnJahr = intervallEndeJahr;
                // intervallEndeJahr = ermittleIntervallEnde(intervallBeginnJahr, IntervallDauer.JAHR);
                long          ausfuehrungsZeitpunktJahr = intervallEndeJahr + ((_minimaleWarteZeitInStunden + 3) * 60 * 60 * 1000);
                AuftragsListe auftragsListe             = AuftragsListe.getInstanz();
                Auftrag       auftrag;

                // Auftrag für die Intervalldaten (TAG)
                auftrag = new Auftrag(ausfuehrungsZeitpunktTag, intervallBeginnTag, intervallEndeTag, vewLZZS, false, IntervallDauer.DTV_TAG);
                auftragsListe.addAuftrag(auftrag);

                // Auftrag für die MONATS-Werte
                auftrag = new Auftrag(ausfuehrungsZeitpunktMonat, intervallBeginnMonat, intervallEndeMonat, vewLZZS, false, IntervallDauer.DTV_MONAT);
                auftragsListe.addAuftrag(auftrag);

                // Auftrag für die Quartals-Werte
                auftrag = new Auftrag(ausfuehrungsZeitpunktQuartal, intervallBeginnQuartal, intervallEndeQuartal, vewLZZS, false, IntervallDauer.DTV_QUARTAL);
                auftragsListe.addAuftrag(auftrag);

                // Auftrag für die JAHRES-Werte
                auftrag = new Auftrag(ausfuehrungsZeitpunktJahr, intervallBeginnJahr, intervallEndeJahr, vewLZZS, false, IntervallDauer.DTV_JAHR);
                auftragsListe.addAuftrag(auftrag);

                // Auftrag für die JAHRES-Werte
                auftrag = new Auftrag(ausfuehrungsZeitpunktJahr, intervallBeginnJahr, intervallEndeJahr, vewLZZS, false, IntervallDauer.MSV_JAHR);
                auftragsListe.addAuftrag(auftrag);
            }
        }

        // Anmeldung des Verwaltungsobjekts und der ATG zum Empfangen
        DataDescription datenBeschreibung;

        // Anmeldung auf Empfang von externen Vorgaben über DAV
        datenBeschreibung = new DataDescription(_konfiguration.getAttributeGroup("atg.aggrLzzsDtvSteuerung"), _konfiguration.getAspect("asp.anfrage"));
        _dav.subscribeReceiver(this, _verwaltungsObjektAggrLzzs, datenBeschreibung, ReceiveOptions.normal(), ReceiverRole.drain());
    }

    /**
     * Initialisierung. Wird durch den {@link de.bsvrz.sys.funclib.application.StandardApplicationRunner} aufgerufen.
     *
     * @param dav Die Datenverteilerverbindung.
     */
    @Override
    public void initialize(ClientDavInterface dav) {
        _dav           = dav;
        _konfiguration = _dav.getDataModel();
    }

    //~ GET METHODEN ==========================================================

    /**
     * Diese Methode muss von der Applikation implementiert werden, um zu signalisieren, ob Sendesteuerungen erwünscht sind
     * und mit der Methode <code>dataRequest</code> verarbeitet werden. In der Implementierung dieser Methode dürfen keine
     * synchronen Aufrufe, die auf Telegramme vom Datenverteiler warten (wie z.B. Konfigurationsanfragen) durchgeführt
     * werden, da ansonsten ein Deadlock entsteht.
     *
     * @param object          Das in der zugehörigen Sendeanmeldung angegebene System-Objekt.
     * @param dataDescription Die in der zugehörigen Sendeanmeldung angegebenen beschreibenden Informationen der
     *                        angemeldeten Daten.
     *
     * @return <code>true</code>, falls Sendesteuerungen gewünscht sind, sonst <code>false</code>.
     *
     * @see #dataRequest
     */
    @Override
    public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
        return false;
    }

    //~ METHODEN ==============================================================

    /**
     * Wertet die Aufrufparameter aus.
     *
     * @param argumentList Aufrufparamter der Applikation
     *
     * @throws Exception Wenn die Aufrufparameter nicht vollständig ausgewertet werden können.
     */
    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        debug = Debug.getLogger();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -pidVerwaltungsObjekt
        // ------------------------------------------------------------------------------------------------------
        _pidVerwaltungsObjektAggrLzzsAufrufParameter = argumentList.fetchArgument("-pidVerwaltungsObjektAggrLzzsDtv=automatisch").asNonEmptyString();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -listePids
        // ------------------------------------------------------------------------------------------------------
        _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=dz.scheppau.a39.49_400").asNonEmptyString();

        // _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=alle").asNonEmptyString();
        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        try {
            _vonZeitPunktAufrufParameter = argumentList.fetchArgument("-vonZeitPunkt=jetzt").asTime();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-vonZeitPunkt]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bisZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        try {
            _bisZeitPunktAufrufParameter = argumentList.fetchArgument("-bisZeitPunkt=jetzt").asTime();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-bisZeitPunkt]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -intervallDauerInStunden
        // ------------------------------------------------------------------------------------------------------
        _intervallDauerInStundenAufrufParameter = argumentList.fetchArgument("-intervallDauerInStunden=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -nachberechnungInTagen
        // ------------------------------------------------------------------------------------------------------
        _nachberechnungInTagenAufrufParameter = argumentList.fetchArgument("-nachberechnungInTagen=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -einmaligeNachBerechnungAufrufParameter
        // ------------------------------------------------------------------------------------------------------
        try {
            _einmaligeNachBerechnungAufrufParameter = argumentList.fetchArgument("-einmaligeNachBerechnung=nein").booleanValue();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-einmaligeNachBerechnung]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -aggregationAusErfasstenDaten
        // ------------------------------------------------------------------------------------------------------
        try {
            _isAggregationAusErfasstenDatenAufrufParameter = argumentList.fetchArgument("-aggregationAusErfasstenDaten=ja").booleanValue();

            // _isAggregationAusErfasstenDatenAufrufParameter = argumentList.fetchArgument("-sggregationAusErfasstenDaten=nein").booleanValue();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-aggregationAusErfasstenDaten]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -qWerteAusReimportAtgVerkehrsDatenLangzeitIntervall
        // ------------------------------------------------------------------------------------------------------
        try {
            _isQWerteAusReimportAtgVerkehrsDatenLangzeitIntervallAufrufParameter = argumentList.fetchArgument("-qWerteAusReimportAtgVerkehrsDatenLangzeitIntervall=ja").booleanValue();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-qWerteAusReimportAtgVerkehrsDatenLangzeitIntervall]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStunden
        // ------------------------------------------------------------------------------------------------------
        _minimaleWarteZeitInStundenAufrufParameter = argumentList.fetchArgument("-minimaleWarteZeitInStunden=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: _grenzenXXX
        // ------------------------------------------------------------------------------------------------------
        _grenzenGeschwindigKeitsKlassenLVOAufrufParameter = argumentList.fetchArgument("-grenzenGeschwindigKeitsKlassenLVO=40:50:60:70:80:90:100:110:120:130:140:150:160:170:180").asNonEmptyString().trim();
        _grenzenGeschwindigKeitsKlassenSGVAufrufParameter = argumentList.fetchArgument("-grenzenGeschwindigKeitsKlassenSGV=40:50:60:70:80:90:100:110:120").asNonEmptyString().trim();
        _grenzenGeschwindigKeitsKlassenBPAAufrufParameter = argumentList.fetchArgument("-grenzenGeschwindigKeitsKlassenBPA=40:50:60:70:80:90:100:110:120").asNonEmptyString().trim();
    }

    // Startet die eigentliche Aggregation, werte die Auftragsliste aus und aggregiert die Daten.
    private void run() {
        debug.info(String.format("%n=================================================================%n" + "Programm Aggregation Langzeitdaten DTV gestartet...%n" + "=================================================================%n"));

        // Auftragsliste holen
        AuftragsListe auftragsListe = AuftragsListe.getInstanz();

        // Anmeldung des Verwaltungsobjekts und der ATG zum Senden
        DataDescription datenBeschreibung;

        // Anmeldung auf das Versenden von Statusinformationen / Fortschrittsanzeige
        datenBeschreibung = new DataDescription(_konfiguration.getAttributeGroup("atg.aggrLzzsDtvStatus"), _konfiguration.getAspect("asp.aggrLzzsDtvStatus"));

        // Datensatz für die Statusinformationen
        Data statusInfo = _dav.createData(_konfiguration.getAttributeGroup("atg.aggrLzzsDtvStatus"));

        try {
            _dav.subscribeSender(this, _verwaltungsObjektAggrLzzs, datenBeschreibung, SenderRole.source());
        }
        catch (OneSubscriptionPerSendData e) {
            debug.error(String.format("Versuch der Doppelanmeldung der Statusanzeige mit Datenbeschreibung [%s]%nProgramm wird beendet.", datenBeschreibung), e);
            System.exit(-1);
        }

        // Endlosschleife mit Prüfung der Auftragsliste.
        // Sobald keinerlei Aufträge mehr vorhanden sind UND der Aufrufparameter -einmaligeNachBerechnung=ja war,
        // wird das Programm beendet
        while (auftragsListe.hatNochAuftraege() ||!_einmaligeNachBerechnung) {
            Auftrag auftrag;
            Auftrag folgeAuftrag;

            if ((auftrag = auftragsListe.holeNaechstenAuftrag()) != null) {
                StringBuffer statusMeldung = new StringBuffer();

                // Ggf. Folgeauftrag in Liste eintragen
                folgeAuftrag = erstelleFolgeAuftrag(auftrag);
                statusMeldung.append(String.format("Aktueller Auftrag: %s", auftrag));

                if (folgeAuftrag != null) {
                    auftragsListe.addAuftrag(folgeAuftrag);
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", folgeAuftrag));
                } else {
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", "kein weiterer Folgeauftrag mehr zu berücksichtigen"));
                }

                debug.info(statusMeldung.toString());

                // Ausgabe Status 10% - Auftrag wird vorbereitet
                statusInfo.getUnscaledValue("FortschrittProzent").set(10);
                statusInfo.getTextValue("StatusMeldung").setText(statusMeldung.toString());

                try {
                    _dav.sendData(new ResultData(_verwaltungsObjektAggrLzzs, datenBeschreibung, System.currentTimeMillis(), statusInfo));
                }
                catch (SendSubscriptionNotConfirmed e) {
                    debug.warning("Versuch des Versands von Statusinformationen ohne Sendefreigabe");
                }

                // Ausgabe Status 20% - Auftrag wird bearbeitet...
                statusMeldung.append(String.format("%nAggregation wird durchgeführt..."));
                statusInfo.getUnscaledValue("FortschrittProzent").set(20);
                statusInfo.getTextValue("StatusMeldung").setText(statusMeldung.toString());

                try {
                    _dav.sendData(new ResultData(_verwaltungsObjektAggrLzzs, datenBeschreibung, System.currentTimeMillis(), statusInfo));
                }
                catch (SendSubscriptionNotConfirmed e) {
                    debug.warning("Versuch des Versands von Statusinformationen ohne Sendefreigabe");
                }

                // Ausgabe Status 100% - Auftrag wurde bearbeitet
                statusMeldung.append(String.format("%nAuftrag durchgeführt:%n%s", auftrag.ausfuehren()));
                statusInfo.getUnscaledValue("FortschrittProzent").set(100);
                statusInfo.getTextValue("StatusMeldung").setText(statusMeldung.toString());

                try {
                    _dav.sendData(new ResultData(_verwaltungsObjektAggrLzzs, datenBeschreibung, System.currentTimeMillis(), statusInfo));
                }
                catch (SendSubscriptionNotConfirmed e) {
                    debug.warning("Versuch des Versands von Statusinformationen ohne Sendefreigabe");
                }

                debug.info(String.format("Ausgeführter Auftrag: %s", auftrag));
            } else {

                // Zurzeit ist nichts zu tun, also warten und es dann noch mal versuchen...
                debug.info(String.format("Zur Zeit keine sofort auszuführenden Aufträge in der Auftragsliste. Warte %d Sekunden.", WARTEZEIT_AUF_NAECHSTEN_AUFTRAG / 1000));

                try {
                    Thread.sleep(WARTEZEIT_AUF_NAECHSTEN_AUFTRAG);
                }
                catch (InterruptedException e) {
                    debug.info("Sleep der Auftragsbearbeitung wurde vorzeitig unterbrochen: ", e);
                }
            }
        }

        debug.info(String.format("%n=================================================================%n" + "Programm Aggregation Langzeitdaten DTV wird ordnungsgemäß beendet,%n" + "da keinerlei Aufträge (weder aktuell noch zukünftig) anliegen%n" + "=================================================================%n"));
        System.exit(0);
    }

    /**
     * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den
     * Datenverteiler-Applikationsfunktionen aufgerufen wird. Diese Methode muss von der Applikation zur Verarbeitung der
     * empfangenen Datensätze implementiert werden.
     *
     * @param results Feld mit den empfangenen Ergebnisdatensätzen.
     */
    @Override
    public void update(ResultData[] results) {
        AuftragsListe auftragsListe = AuftragsListe.getInstanz();

        // Ergebnisdatensatz mit zusätzlichen Einmalaufträgen auswerten und in die Auftragsliste eintragen
        for (ResultData ergebnis : results) {
            if (ergebnis.hasData()) {
                Data         daten              = ergebnis.getData();
                long         von                = daten.getTimeValue("von").getMillis();
                long         bis                = daten.getTimeValue("bis").getMillis();
                int          anz                = daten.getArray("Lzzs").getLength();
                boolean      berechneDTVTag     = daten.getTextValue("berechneDTVTag").getText().equals("Ja");
                boolean      berechneDTVMonat   = daten.getTextValue("berechneDTVMonat").getText().equals("Ja");
                boolean      berechneDTVQuartal = daten.getTextValue("berechneDTVQuartal").getText().equals("Ja");
                boolean      berechneDTVJahr    = daten.getTextValue("berechneDTVJahr").getText().equals("Ja");
                boolean      berechneMSVJahr    = daten.getTextValue("berechneMSVJahr").getText().equals("Ja");
                List<String> listeLzzs          = new ArrayList<String>();

                if (anz == 0) {

                    // Für alle Lzzs die Aufträge eintragen
                    listeLzzs = _listePidsLzzs;
                } else {
                    for (int i = 0; i < anz; i++) {
                        listeLzzs.add(daten.getArray("Lzzs").getItem(i).asReferenceValue().getSystemObjectPid());
                    }
                }

                // Intervallzeiträume ermitteln...
                // von und bis auf korrekten Intervallanfang und Intervallende setzten, Anzahl der Intervalle berechnen
                // =============================================================================================
                // Zeitpunkte für den TAG bestimmen
                // =============================================================================================
                if (berechneDTVTag) {
                    Calendar intervallBeginnTag = new GregorianCalendar();
                    Calendar intervallEndeTag   = new GregorianCalendar();

                    intervallBeginnTag.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_TAG, 0));
                    intervallEndeTag.setTimeInMillis(ermittleIntervallEnde(bis, IntervallDauer.DTV_TAG));

                    long anzIntervalleTag = (intervallEndeTag.getTimeInMillis() - intervallBeginnTag.getTimeInMillis()) / (24 * 60 * 60 * 1000);

                    // Aufträge eintragen (für alle Lzzs und alle Intervalle)
                    for (String pid : listeLzzs) {
                        VewLZZS vewLZZS = _mapVewLZZS.get(pid);

                        if (vewLZZS != null) {
	                        // Intervallbeginn auf den in der ATG gewünschten Zeitpunkt zurücksetzen,
	                        // da der Wert nachfolgend hochgezählt wird --> beim nächster LZZS wäre er deshalb falsch
	                        intervallBeginnTag.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_TAG, 0));
                            for (int i = 0; i < anzIntervalleTag; i++) {

                                // Intervallbeginn
                                long intervallVon = intervallBeginnTag.getTimeInMillis();

                                // Ein Intervall weiter (als Intervallende, im nächsten Zyklus als Intervallbeginn)
                                intervallBeginnTag.add(Calendar.DAY_OF_YEAR, 1);

                                long    intervallBis = intervallBeginnTag.getTimeInMillis();
                                Auftrag auftrag      = new Auftrag(intervallBis + _minimaleWarteZeitInStunden * 60 * 60 * 1000, intervallVon, intervallBis, vewLZZS, true, IntervallDauer.DTV_TAG);

                                auftragsListe.addAuftrag(auftrag);
                                debug.info(String.format("Auftrag aus externer Vorgabe zur Auftragsliste hinzugefügt: %s%n", auftrag));
                            }
                        }
                    }
                }

                // =============================================================================================
                // Zeitpunkte für den Monat bestimmen
                // =============================================================================================
                if (berechneDTVMonat) {
                    Calendar intervallBeginnMonat = new GregorianCalendar();
                    Calendar intervallEndeMonat   = new GregorianCalendar();

                    intervallBeginnMonat.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_MONAT, 0));
                    intervallEndeMonat.setTimeInMillis(ermittleIntervallEnde(bis, IntervallDauer.DTV_MONAT));

                    long anzIntervalleMonat = ermittleAnzMonate(intervallBeginnMonat, intervallEndeMonat);

                    // Aufträge eintragen (für alle Lzzs und alle Intervalle)
                    for (String pid : listeLzzs) {
                        VewLZZS vewLZZS = _mapVewLZZS.get(pid);

                        if (vewLZZS != null) {
	                        // Intervallbeginn auf den in der ATG gewünschten Zeitpunkt zurücksetzen,
	                        // da der Wert nachfolgend hochgezählt wird --> beim nächster LZZS wäre er deshalb falsch
	                        intervallBeginnMonat.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_MONAT, 0));
                            for (int i = 0; i < anzIntervalleMonat; i++) {

                                // Intervallbeginn
                                long intervallVon = intervallBeginnMonat.getTimeInMillis();

                                // Ein Intervall weiter (als Intervallende, im nächsten Zyklus als Intervallbeginn)
                                intervallBeginnMonat.add(Calendar.MONTH, 1);

                                long    intervallBis = intervallBeginnMonat.getTimeInMillis();
                                Auftrag auftrag      = new Auftrag(intervallBis + ((_minimaleWarteZeitInStunden + 1) * 60 * 60 * 1000), intervallVon, intervallBis, vewLZZS, true, IntervallDauer.DTV_MONAT);

                                auftragsListe.addAuftrag(auftrag);
                                debug.info(String.format("Auftrag aus externer Vorgabe zur Auftragsliste hinzugefügt: %s%n", auftrag));
                            }
                        }
                    }
                }

                // =============================================================================================
                // Zeitpunkte für das Quartal bestimmen
                // =============================================================================================
                if (berechneDTVQuartal) {
                    Calendar intervallBeginnQuartal = new GregorianCalendar();
                    Calendar intervallEndeQuartal   = new GregorianCalendar();

                    intervallBeginnQuartal.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_QUARTAL, 0));
                    intervallEndeQuartal.setTimeInMillis(ermittleIntervallEnde(bis, IntervallDauer.DTV_QUARTAL));

                    long anzIntervalleQuartal = ermittleAnzMonate(intervallBeginnQuartal, intervallEndeQuartal) / 3;

                    // Aufträge eintragen (für alle Lzzs und alle Intervalle)
                    for (String pid : listeLzzs) {
                        VewLZZS vewLZZS = _mapVewLZZS.get(pid);

                        if (vewLZZS != null) {
	                        // Intervallbeginn auf den in der ATG gewünschten Zeitpunkt zurücksetzen,
	                        // da der Wert nachfolgend hochgezählt wird --> beim nächster LZZS wäre er deshalb falsch
	                        intervallBeginnQuartal.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_QUARTAL, 0));
                            for (int i = 0; i < anzIntervalleQuartal; i++) {

                                // Intervallbeginn
                                long intervallVon = intervallBeginnQuartal.getTimeInMillis();

                                // Ein Intervall weiter (als Intervallende, im nächsten Zyklus als Intervallbeginn)
                                intervallBeginnQuartal.add(Calendar.MONTH, 3);

                                long    intervallBis = intervallBeginnQuartal.getTimeInMillis();
                                Auftrag auftrag      = new Auftrag(intervallBis + ((_minimaleWarteZeitInStunden + 2) * 60 * 60 * 1000), intervallVon, intervallBis, vewLZZS, true, IntervallDauer.DTV_QUARTAL);

                                auftragsListe.addAuftrag(auftrag);
                                debug.info(String.format("Auftrag aus externer Vorgabe zur Auftragsliste hinzugefügt: %s%n", auftrag));
                            }
                        }
                    }
                }

                // =============================================================================================
                // Zeitpunkte für das Jahr bestimmen
                // =============================================================================================
                if (berechneDTVJahr) {
                    Calendar intervallBeginnJahr = new GregorianCalendar();
                    Calendar intervallEndeJahr   = new GregorianCalendar();

                    intervallBeginnJahr.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_JAHR, 0));
                    intervallEndeJahr.setTimeInMillis(ermittleIntervallEnde(bis, IntervallDauer.DTV_JAHR));

                    long anzIntervalleJahr = intervallEndeJahr.get(Calendar.YEAR) - intervallBeginnJahr.get(Calendar.YEAR);

                    // Aufträge eintragen (für alle Lzzs und alle Intervalle)
                    for (String pid : listeLzzs) {
                        VewLZZS vewLZZS = _mapVewLZZS.get(pid);

                        if (vewLZZS != null) {
	                        // Intervallbeginn auf den in der ATG gewünschten Zeitpunkt zurücksetzen,
	                        // da der Wert nachfolgend hochgezählt wird --> beim nächster LZZS wäre er deshalb falsch
	                        intervallBeginnJahr.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.DTV_JAHR, 0));
                            for (int i = 0; i < anzIntervalleJahr; i++) {

                                // Intervallbeginn
                                long intervallVon = intervallBeginnJahr.getTimeInMillis();

                                // Ein Intervall weiter (als Intervallende, im nächsten Zyklus als Intervallbeginn)
                                intervallBeginnJahr.add(Calendar.YEAR, 1);

                                long    intervallBis = intervallBeginnJahr.getTimeInMillis();
                                Auftrag auftrag      = new Auftrag(intervallBis + ((_minimaleWarteZeitInStunden + 3) * 60 * 60 * 1000), intervallVon, intervallBis, vewLZZS, true, IntervallDauer.DTV_JAHR);

                                auftragsListe.addAuftrag(auftrag);
                                debug.info(String.format("Auftrag aus externer Vorgabe zur Auftragsliste hinzugefügt: %s%n", auftrag));
                            }
                        }
                    }
                }

                // =============================================================================================
                // Zeitpunkte für das Jahr MSV bestimmen
                // =============================================================================================
                if (berechneMSVJahr) {
                    Calendar intervallBeginnJahr = new GregorianCalendar();
                    Calendar intervallEndeJahr   = new GregorianCalendar();

                    intervallBeginnJahr.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.MSV_JAHR, 0));
                    intervallEndeJahr.setTimeInMillis(ermittleIntervallEnde(bis, IntervallDauer.MSV_JAHR));

                    long anzIntervalleJahr = intervallEndeJahr.get(Calendar.YEAR) - intervallBeginnJahr.get(Calendar.YEAR);

                    // Aufträge eintragen (für alle Lzzs und alle Intervalle)
                    for (String pid : listeLzzs) {
                        VewLZZS vewLZZS = _mapVewLZZS.get(pid);

                        if (vewLZZS != null) {
	                        // Intervallbeginn auf den in der ATG gewünschten Zeitpunkt zurücksetzen,
	                        // da der Wert nachfolgend hochgezählt wird --> beim nächster LZZS wäre er deshalb falsch
	                        intervallBeginnJahr.setTimeInMillis(ermittleIntervallBeginn(von, IntervallDauer.MSV_JAHR, 0));
                            for (int i = 0; i < anzIntervalleJahr; i++) {

                                // Intervallbeginn
                                long intervallVon = intervallBeginnJahr.getTimeInMillis();

                                // Ein Intervall weiter (als Intervallende, im nächsten Zyklus als Intervallbeginn)
                                intervallBeginnJahr.add(Calendar.YEAR, 1);

                                long    intervallBis = intervallBeginnJahr.getTimeInMillis();
                                Auftrag auftrag      = new Auftrag(intervallBis + ((_minimaleWarteZeitInStunden + 3) * 60 * 60 * 1000), intervallVon, intervallBis, vewLZZS, true, IntervallDauer.MSV_JAHR);

                                auftragsListe.addAuftrag(auftrag);
                                debug.info(String.format("Auftrag aus externer Vorgabe zur Auftragsliste hinzugefügt: %s%n", auftrag));
                            }
                        }
                    }
                }
            }
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 08.12.10
